/*
* $Id: servo1.c,v 1.2 2002/04/17 21:26:14 ldrolez Exp $
* $Id: NextStep.c, v 1.2 2003/03/14
*
* Danny H. Newport
* 3-2-03
* NextStep clone for avr 2313 with 10mhz clock
* for FirstBot Robot
*
* compiled with AvrEdit Ver. 3.5 using 61% memory (room for expansion)
*
*      Port layout:
*
*      PORTD     0	Uart receive         //not used
*      PORTD     1	Uart send            //not needed
*      PORTD     2      Left whisker switch
*      PORTD     3      Right whisker switch
*      PORTD     4      Left LED IRPD        //not needed
*      PORTD     5      Right LED IRPD       //not needed
*      PORTD     6      IRPD SENSE
*
*      PORTB     0      Left Servo wheel
*      PORTB     1      Right Servo wheel
*      PORTB     2-5    unused PWM outputs
*      PORTB     6      Left LED
*      PORTB     7      Right LED
*
*
*   This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.       
* 
*/

#include <io.h>
#include <interrupt.h>
#include <sig-avr.h>
#include <timer.h>
#define __AVR_AT90S2313__

/* timer 0 ini tial count */
/* 0 -> 25,6 us PWM resolution @ 10 Mhz */
/* 56 -> 20 us */
#define STARTCNT0 56

//defines for delay
#define CPU_freq   10000000                     //cpu clock frequency
#define CYCLES_PER_US ((CPU_freq+500000)/1000000) 	// cpu cycles per microsecond
#undef ECHO

/* global variables for robot functions */
static uint8_t mov = 0; 
static uint16_t m_left,m_right,m_back;
static uint8_t dir = 0;

/* global variables for delay functions */
uint16_t delay_count=CPU_freq/4000;             // used in delay function

//static uint8_t dloop = 0;                     // used for delay loops greater than 255mS

/* global variables used by PWM outputs */
volatile uint8_t pwm[8];
static uint8_t pwm_clk = 0;
static uint8_t pwm_cnt = 0;
volatile uint8_t pwm_config;

/* global variables for serial port */
volatile uint8_t ready = 0;

/* prototypes */
void Moveservo(uint8_t mov);
int read_whiskers(void);
void setdirection(void);
void main_loop (void);
int read_irpd(void);
void delay_ms(uint8_t ms);

//enumerate the servo directions
enum {
		rSTOP             = 0,
		rFORWARD          = 1,
		rBACKWARDS        = 2,
		rLEFT             = 3,
		rRIGHT            = 4,
		rVLEFT            = 5,
		rVRIGHT           = 6
};


/*
* Set default values
*/
void init(void)
{
	/* all outputs = pwm */
	pwm_config = 0xFF;
	/* all outputs low */
	//memset (pwm, 0x00, 8);

	pwm[6]=0;               //LED output for "debugging"
        pwm[7]=0;               //LED output for "debugging"
	m_left=m_right=m_back=0;//clear direction delays
}

/*
* uart transmit complete isr
*/
INTERRUPT (SIG_UART_TRANS)
{
	ready = 1;
}

/*
* Timer isr
*
* Takes 95 cycles (9,5us @10Mhz) with gcc 3.0.3 and -Os optimization
*/
SIGNAL (SIG_OVERFLOW0)
{
	uint8_t tpwm;
	
	/* reload counter, +23 to compensate for all 
	instructions executed between the timer
	overflow and this reload */
	outp (STARTCNT0+23, TCNT0);	
	tpwm = 0;
	if (pwm_clk < pwm[0])
		tpwm |= 1;
	if (pwm_clk < pwm[1])
		tpwm |= 2;
	if (pwm_clk < pwm[2])
		tpwm |= 4;
	if (pwm_clk < pwm[3])
		tpwm |= 8;
	if (pwm_clk < pwm[4])
		tpwm |= 16;
	if (pwm_clk < pwm[5])
		tpwm |= 32;
	if (pwm_clk < pwm[6])
		tpwm |= 64;
	if (pwm_clk < pwm[7])
		tpwm |= 128;
	
	if (pwm_cnt & 1) {		/* in servo mode, only output 1 pulse / 2 */
		tpwm &= pwm_config;
		__asm__ __volatile__ ("nop"); /* improves timing */
	} else {
		__asm__ __volatile__ ("nop");
	}
	
	outp (tpwm, PORTB);		/* set PWM outputs */
	
	pwm_clk++;
	if (pwm_clk == 255)
	{
		pwm_clk = 0;		/* 255 steps not 256 */
		pwm_cnt++;
	}
}

/*
* Convert an ASCII char to a 4 bits value
*/
uint8_t quartet (char c)
{
	/* to upper case */
	if (c >= 'a')
		c = c - 32;
	if (c < 'A')
		return (c - '0');
	else
		return (c - 'A' + 10);
}

char tobcd (uint8_t i)
{
	if (i > 9)
		return (i + 'A' - 10);
	else
		return (i + '0');
}

/*
*  wait for serial buffer to clear
*/
void pause (void)
{
	while (!ready);
	ready = 0;
}


/*
* 
*/
int
main (void)
{
	init();
	/* 10Mhz / 200 = 1 overflow every 20,0 us @ 10MHz */
	/* 255 steps = 5.1 ms */
	outp (0xff, DDRB);		/* portb = output */
	outp (0x00, DDRD);		/* portd = input */
	
	outp ((1 << TOIE0), TIMSK);
	outp (STARTCNT0, TCNT0);	/* timer start val */
	outp (CK, TCCR0);		/* prescale = 1 */
	
	outp ((1 << TXCIE) | (1 << RXEN) | (1 << TXEN), UCR);
	outp (64, UBRR);		/* baud rate =9600 */
	
	sei ();			/* enable interrupts */
	main_loop ();
	return 0;
}

/*
 *  function to get data from whisker switches
 */
int read_whiskers(void)
{
	dir=0;
	//check for wisker switches hitting

        //check left whisker switch
	if(bit_is_set (PIND,PIND2)==0)
	{
		pwm[6]=0;
		dir +=1;
		m_left=75;
		m_back=100;
	}
	else pwm[6]=255;

        //check right whisker switch
	if(bit_is_set (PIND,PIND3)==0)
	{
		pwm[7]=0;
		dir+=2;
		m_right=75;
		m_back=100;
	}else pwm[7]=255;

        //setup for both switches hitting
	if(dir==3)
	{
		m_right=0;
		m_left=0;
		m_back=100;
	}
       
	return dir;
}

/*
 *  function to get data from IR sensors
 */
int read_irpd(void)
{

	//default to forward motion
	dir=0;
	
	//setup  to read left led
        sbi(PORTD,PIND4);
	delay_ms(1);
	cbi(PORTD,PIND4);
	delay_ms(1);
	sbi(PORTD,PIND4);
	delay_ms(1);
	//check for left LED bounce
	if(bit_is_clear (PIND,PIND6)==1)
	{
		pwm[6]=0;
		dir+=1;
		m_left=100;
		m_back=0;
	}
	else pwm[6]=255;
	cbi(PORTD,PIND4);

	//setup to read right led
	sbi(PORTD,PIND5);
	delay_ms(1);
	cbi(PORTD,PIND5);
        delay_ms(1);
        sbi(PORTD,PIND5);
	delay_ms(1);
        //check for right LED bounce
	if(bit_is_clear (PIND,PIND6)==1)
	{
		pwm[7]=0;
		dir+=2;
		m_right=100;
		m_back=0;
	}
	else pwm[7]=255;

	cbi(PORTD,PIND5);

	//check for both leds reading an object
	if(dir==3)
	{
		m_right=0;
		m_left=0;
		m_back=100;
	}
	
	return dir;
}

/*
 *  function to set servo wheel motion
 */
void setdirection(void)
{
	switch(dir)//select direction
	{
	case 0:
		Moveservo(rFORWARD);//forward
		outp('F',UDR);
		break;
	case 1://rxd
		Moveservo(rRIGHT);//right turn
                outp('R',UDR);
		delay_ms(m_left);//remain turning
                m_left=0;        //cancel turn delay
                dir=0;           //cancel right turn
		break;
	case 2://txd
		Moveservo(rLEFT);//left turn
                outp('L',UDR);
		delay_ms(m_right);//remain turning
                m_right=0;        //cancel turn delay
                dir=0;            //cancel left turn
		break;
	case 3:
		Moveservo(rBACKWARDS);//backup
		delay_ms(m_back);
                m_back=0;
                if(m_right==0&&m_left==0)
                        dir=0;
                if(m_left>0)
                        dir=1;
                if(m_right>0)
                        dir=2;
		outp('B',UDR);
		break;
	default:                              //should never get here
		Moveservo(rFORWARD);          //forward
                dir=0;                        //correct direction variable
                m_right=m_back=m_left=0;      //reset directions
		outp('?',UDR);
		break;
		
	}
}

/*
 *  function for endless loop
 */
void main_loop (void)
{
	dir = 0; //forward direction
	mov = 1; //forward
	
	outp(0x00,DDRD);       //setup port D as input

	for (;;)		//loop forever
	{

		if(dir==0)//check for forward moving direction
		{
			dir = read_whiskers();       //read whisker switches
			if(dir==0)                   //if not hitting whisker
				dir = read_irpd();   //then read irpd
		}
		
		//double check directions
                if(m_back>0)
                        dir=3;
                else if(m_left>0)
                        dir=1;
                else if(m_right>0)
                        dir=2;
                else dir=0;

                //command robot direction
		setdirection();
		
	}
}


/*
 *  function to set wheel pwm
 */
void Moveservo(uint8_t mov)
{
	switch (mov)
	{
	case 0://stop
		pwm[0] = 0;	/* set pwm */
		pwm[1] = 0 ;	/* set pwm */
		break;
	case 1://forward
		pwm[0] = 85;	/* set pwm */
		pwm[1] = 20;	/* set pwm */
		break;
	case 2://backwards
		pwm[0] = 20;	/* set pwm */
		pwm[1] = 85;	/* set pwm */
		break;
	case 3://left
		pwm[0] = 20;	/* set pwm */
		pwm[1] = 20;	/* set pwm */
		break;
	case 4://right
		pwm[0] = 85;	/* set pwm */
		pwm[1] = 85;	/* set pwm */
		break;
		
	}
	
}
/*void delay_ms(char delay_time)
{
	int c;
	for(c = 0; c < (CPU_freq / 40000); c++)
	{
	}
}*/

/*// microsecond delay function
void delay_us(unsigned short us)
{
    unsigned short  delay_loops;
    register unsigned short  i;

    delay_loops = (us+3)/5*CYCLES_PER_US; // +3 for rounding up (dirty)

	// one loop takes 5 cpu cycles
    for (i=0; i < delay_loops; i++) {};
}*/

// millisecond delay function
void delay_ms(uint8_t ms)
{
        uint16_t cnt; asm volatile (

        "\n" "L_dl1%=:" "\n\t" "mov %A0, %A2" "\n\t" "mov %B0, %B2" "\n" "L_dl2%=:" "\n\t"

        "sbiw %A0, 1" "\n\t" "brne L_dl2%=" "\n\t" "dec %1" "\n\t" "brne L_dl1%=" "\n\t" : "=&w" (cnt) : "r" (ms), "r" (delay_count) );
}

